using Unity.Transforms;
using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using Unity.Entities;
using Unity.Mathematics;
using Unity.Collections;
using Unity.Jobs;
using Unity.Burst;

[System.Serializable]
public struct TrailCompleteRenderer : IComponentData
{
	public float3 currentPosition;
	public float progress;
	public float speed;
	public bool first;
	public bool teleport;
	public double lastEnableTime;
}
[RequiresEntityConversion]
public class TrailCompleteRendererAuthoring: MonoBehaviour, IConvertGameObjectToEntity
{
    public TrailCompleteRenderer trailCompleteRenderer;
    public void Convert(Entity entity, EntityManager dstManager, GameObjectConversionSystem conversionSystem)
    {
        dstManager.AddComponentData(entity, trailCompleteRenderer);
		GetComponent<TrailRenderer>().enabled = false;
    }
}

public class TrailCompleteRenderSystem : ComponentSystem
{
	protected override void OnUpdate()
	{
		Entities.ForEach((TrailRenderer trail,  ref TrailCompleteRenderer renderer) =>
		{
			var newPos = renderer.currentPosition;
			newPos.y = trail.transform.position.y;
			if (renderer.teleport)
			{
				trail.enabled = false;
				trail.transform.position = newPos;
				renderer.teleport = false;
				renderer.first = true;
				renderer.lastEnableTime = Time.ElapsedTime;
			}
			if (renderer.first)
			{
				trail.enabled = true;
				renderer.first = false;
			}
			
			trail.transform.position = newPos;
		});
	}
}

public class TrailCompletePositionJobSystem : JobComponentSystem
{	
	EntityQuery trailDataQuery;
	EntityQuery trailLoopQuery;
	EntityQuery trailRenderQuery;
	BeginSimulationEntityCommandBufferSystem bufferSystem;
	protected override void OnCreate()
	{
		base.OnCreateManager();
		trailLoopQuery = GetEntityQuery(ComponentType.ReadOnly<TrailLoop>(), ComponentType.Exclude<TrailFinished>());
		trailRenderQuery = GetEntityQuery(ComponentType.ReadWrite<TrailCompleteRenderer>());

		bufferSystem = World.GetOrCreateSystem<BeginSimulationEntityCommandBufferSystem>();
	}
	
	protected override JobHandle OnUpdate(JobHandle inputDeps)
	{
		JobHandle TrailLoopHandle, TrailRenderHandle, TrailRenderEntityHandle, TrailLoopPointEntitiesHandle ;
		var TrailLoopArray = trailLoopQuery.ToComponentDataArray<TrailLoop>(Allocator.TempJob, out TrailLoopHandle);
		var TrailLoopPointEntities = trailLoopQuery.ToEntityArray(Allocator.TempJob, out TrailLoopPointEntitiesHandle);
		var TrailRenderARray = trailRenderQuery.ToComponentDataArray<TrailCompleteRenderer>(Allocator.TempJob, out TrailRenderHandle);
		var TrailEntities = trailRenderQuery.ToEntityArray(Allocator.TempJob, out TrailRenderEntityHandle);

		var dep = JobHandle.CombineDependencies(inputDeps,  TrailLoopHandle );
		dep = JobHandle.CombineDependencies(dep, TrailRenderHandle, TrailRenderEntityHandle);
		dep = JobHandle.CombineDependencies(dep, TrailLoopPointEntitiesHandle);
		var job = new Job
		{
			 TrailLoopPointEntities = TrailLoopPointEntities
			, TrailLoopArray = TrailLoopArray
			, TrailRenderArray = TrailRenderARray
			, deltaTime = Time.DeltaTime
			, cmdBuffer = bufferSystem.CreateCommandBuffer().ToConcurrent()
			, TrailRenderEntityArray = TrailEntities
			, pointBuffers = GetBufferFromEntity<TrailLoopPoint>(true)

		}.Schedule(dep);
		job.Complete();
		//bufferSystem.AddJobHandleForProducer(dep);
		return job;
	}

	public struct Job : IJob
	{
		[DeallocateOnJobCompletion]
		public NativeArray<Entity> TrailLoopPointEntities;


		[DeallocateOnJobCompletion]
		public NativeArray<TrailLoop> TrailLoopArray;
		[DeallocateOnJobCompletion]
		public NativeArray<TrailCompleteRenderer> TrailRenderArray;
		[DeallocateOnJobCompletion]
		public NativeArray<Entity> TrailRenderEntityArray;

		public BufferFromEntity<TrailLoopPoint> pointBuffers;

		public EntityCommandBuffer.Concurrent cmdBuffer;

		public float deltaTime;		
		public void Execute()
		{
			for (int i = 0; i < TrailLoopArray.Length; i++)
			{
				var loop = TrailLoopArray[i];
				var loopPointEntity = TrailLoopPointEntities[i];
				var dynBuffer = pointBuffers[loopPointEntity];

				int trailRenderIndex = loop.trailIndex;
				bool renderNeedsTeleport = false;
				//find the best trail renderer.
				if (trailRenderIndex ==-1)
				{
					int bestIndex = 0;
					double bestEnableTime = TrailRenderArray[0].lastEnableTime;
					for (int j=1;j<TrailRenderArray.Length;j++)
					{
						if (TrailRenderArray[j].lastEnableTime < bestEnableTime)
						{
							bestIndex = j;
							bestEnableTime = TrailRenderArray[j].lastEnableTime;
						}
					}
					trailRenderIndex = bestIndex;
					loop.trailIndex = trailRenderIndex;
					cmdBuffer.SetComponent(0, TrailLoopPointEntities[i], loop);
					renderNeedsTeleport = true;
				}

				var render = TrailRenderArray[trailRenderIndex];
				if (renderNeedsTeleport)
				{
					render.teleport = true;
					render.progress = 0.0f;
				}
				render.progress += deltaTime * render.speed;
				var cumulativeDistance = 0.0f;
				bool foundEnd = false;
				for(int j=1;j<dynBuffer.Length;j++)
				{
					var thisPosition = dynBuffer[j].position;
					var previousPosition = dynBuffer[j - 1].position;
					var direction = thisPosition - previousPosition;
					var distance = math.length(direction);
					
					if (cumulativeDistance+distance > render.progress)
					{
						direction = math.normalizesafe(direction);
						var positionProgress = render.progress - cumulativeDistance;
						render.currentPosition = previousPosition + direction * positionProgress;
						render.first = true;
						foundEnd = true;
						break;
					}
					else
					{
						cumulativeDistance += distance;
					}
				}
				//if we haven't found the end, we must be at the last position.
				if (!foundEnd)
				{
					render.currentPosition = dynBuffer[dynBuffer.Length - 1].position;
					render.first = true;
					foundEnd = true;
					float3 midPoint = float3.zero;
					for (int j = 0; j < dynBuffer.Length; j++)
					{
						midPoint += dynBuffer[j].position;
					}
					cmdBuffer.AddComponent<TrailFinished>(0, loopPointEntity, new TrailFinished() { midPoint = midPoint / dynBuffer.Length });
				}
				cmdBuffer.SetComponent(0,TrailRenderEntityArray[trailRenderIndex], render);
			}
		}
	}
}