using Unity.Transforms;
using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using Unity.Entities;
using Unity.Mathematics;
using Unity.Collections;
using Unity.Jobs;
using Unity.Burst;

public class TrailLinkingSystem : ComponentSystem
{
	int runningId = 0;
	protected override void OnUpdate()
	{
		
		
		Entities.ForEach((Entity ent, ref Player player, ref Translation translation, ref TrailLayer trailLayer) =>
		{
			var distanceFromLastPoint = trailLayer.lastSpawnPoint - translation.Value;

			bool isOverMaxDistance = math.lengthsq(distanceFromLastPoint) > trailLayer.maxDistance * trailLayer.maxDistance;
			bool isOverMinDistance = math.lengthsq(distanceFromLastPoint) > trailLayer.minDistance * trailLayer.minDistance;
			bool isOvertime = Time.ElapsedTime > trailLayer.lastSpawnTime + trailLayer.timeBetweenSpawns;
			if ( (isOverMinDistance && isOvertime) || isOverMaxDistance)
			{
				trailLayer.lastSpawnTime = Time.ElapsedTime;

				var position = translation.Value;
				
				var newEntity = PostUpdateCommands.CreateEntity();
				PostUpdateCommands.AddComponent(newEntity, new TrailSegment() { position_WS = translation.Value, previousPosition_WS = trailLayer.lastSpawnPoint,time = Time.ElapsedTime, id = runningId++ });
				trailLayer.lastSpawnPoint = position;
			}
		});

		Entities.ForEach((ref TrailSegment trailData) =>
		{
			Debug.DrawLine(trailData.previousPosition_WS, trailData.position_WS);
		});
	}
}

public class TrailCheck : JobComponentSystem
{	
	EntityQuery segmentQuery;

	BeginInitializationEntityCommandBufferSystem bufferSystem;
	NativeList<float3> trailPoints;
	protected override void OnCreate()
	{
		base.OnCreateManager();
		segmentQuery = GetEntityQuery( ComponentType.ReadWrite<TrailSegment>(), ComponentType.Exclude<TrailLoopTag>());
		bufferSystem = World.GetOrCreateSystem<BeginInitializationEntityCommandBufferSystem>();
		trailPoints = new NativeList<float3>(1000,Allocator.Persistent);
	}

	protected override void OnDestroy()
	{
		base.OnDestroy();
		trailPoints.Dispose();
	}
	protected override JobHandle OnUpdate(JobHandle inputDeps)
	{
		trailPoints.Clear();

		JobHandle  TrailDataHandle ;
		JobHandle gatherEntities;
		var segmentDataArry = segmentQuery.ToComponentDataArray<TrailSegment>(Allocator.TempJob, out TrailDataHandle);
		
		var segmentEntityArray = segmentQuery.ToEntityArray(Allocator.TempJob, out gatherEntities);
		var dep = JobHandle.CombineDependencies(inputDeps,  TrailDataHandle );
		dep = JobHandle.CombineDependencies(dep,gatherEntities);
		NativeArray<TrailLoop> trailLoop = new NativeArray<TrailLoop>(1, Allocator.TempJob);
		trailLoop[0] = new TrailLoop();
		var job = new LoopJob
		{
			TrailDataArray = segmentDataArry,
			trailArray = trailLoop,
			trailPoints = trailPoints

		}.Schedule(dep);

		var completeJob = new LoopCompleteJob
		{
			entityArray = segmentEntityArray,
			trailArray = trailLoop,
			cmdBuffer = bufferSystem.CreateCommandBuffer(),
			trailPoints = trailPoints
		}.Schedule(job);

		return completeJob;
	}

	public struct LoopCompleteJob : IJob
	{

		public EntityCommandBuffer cmdBuffer;
		[DeallocateOnJobCompletion]
		public NativeArray<TrailLoop> trailArray;

		[DeallocateOnJobCompletion]
		public NativeArray<Entity> entityArray;

		public NativeList<float3> trailPoints;
		public void Execute()
		{
			var newTrailLoop = trailArray[0];
			if (newTrailLoop.startId != newTrailLoop.endId)
			{
				var newEntity = cmdBuffer.CreateEntity();
				cmdBuffer.AddComponent(newEntity, newTrailLoop);
				var buffer = cmdBuffer.AddBuffer<TrailLoopPoint>(newEntity);
				foreach(var point in trailPoints)
				{
					buffer.Add(new TrailLoopPoint() { position = point });
				}
				for (int i=0;i<entityArray.Length;i++)
				{
					cmdBuffer.AddComponent(entityArray[i], new TrailLoopTag());
				}
			}
		}
	}

	public struct LoopJob : IJob
	{		

		[DeallocateOnJobCompletion]
		public NativeArray<TrailSegment> TrailDataArray;


		public NativeArray<TrailLoop> trailArray;
		public NativeList<float3> trailPoints;
		public void Execute()
		{
			NativeSortExtension.Sort(TrailDataArray);
			for (int i = 0; i < TrailDataArray.Length/2; i++)
			{
				var line1 = TrailDataArray[i];
				for (int j= TrailDataArray.Length-1;j>i+1;j--)
				{					
					var line2 = TrailDataArray[j];
					float3 intersection;

					bool intersects = LineHelpers.FindIntersection(line1.previousPosition_WS, line1.position_WS, line2.previousPosition_WS, line2.position_WS, out intersection);
					if (intersects)
					{
						TrailLoop newTrailLoop = new TrailLoop() { startId = math.min(line1.id, line2.id), endId = math.max(line1.id, line2.id), trailIndex = -1 };

						intersection.y = 0.0f;
						trailPoints.Add(intersection);
						for (int k=i;k<=j;k++)
						{
							var addingTrailData = TrailDataArray[k];
							// if it's the first one, add the previous point
							if (k==i)
							{
								trailPoints.Add(addingTrailData.previousPosition_WS);
							}
							trailPoints.Add(addingTrailData.position_WS);
						}

						trailPoints.Add(intersection);
						trailPoints.Add(trailPoints[0]);
						trailArray[0] = newTrailLoop;
						return;
					}
				}
			}
		}



	}
}